# archiver.py
#
# Project: AutoArchive
# License: GNU GPLv3
#
# Copyright (C) 2003 - 2009 Robert Cernansky
# Copyright (C) 2003 - 2009 HOME software



"""Provides Archiver and derived classes.
"""



# INCLUDE FILES

import subprocess
import glob
import shutil

from aautils import *
from options import options



# CONSTANTS

# enumerates error codes returned with ExecuteException
ErrorCodes = Enum(

    # error code not specified
    "Undefined",

    # non fatal error occured
    "Warning",

    # error occured, operation was not successfull
    "Error")



# CLASSES

class Archiver(object):
    "Abstract class for an archiver program."

    # full path to archiver program
    ARCHIVER = None

    # suffix (extension) of files created by ARCHIVER
    _SUFFIX = None

    def __init__(self, archiveSpec):

        # raise exception because this class is meant as abstract
        if not self.ARCHIVER:
            raise NotImplementedError
        
        self.archiveSpec = archiveSpec
        self.level = self.__getLevel()
        self.archiveFileName = self.getArchiveFileName()



    def execute(self):
        """Execute archiver program to create archive from archiveSpec.

        Raises ExecuteException if some error occures. The error code is in
        errorCode attribute."""

        arguments = self.arguments()
        environment = os.environ.copy()
        environment.update(arguments[1])
        try:
            subprocess.check_call([self.ARCHIVER] + arguments[0],
                                  env = environment)
        except OSError, ex:
            raise ExecuteException, (str(ex), ErrorCodes.Error)
        except subprocess.CalledProcessError, ex:
            raise ExecuteException, (self._getErrorMessage(ex.returncode),
                                     self._getErrorCode(ex.returncode))



    def arguments(self):
        """Converts archiveSpec and options to form suitable for the archiver.

        Returns tuple (options, environment)."""

        return ([], None)



    def getArchiveFileName(self):
        "Returns name for archive file."

        return self.archiveSpec.name + self._getLevelSuffix(self.level) + \
            self.suffix



    def _getLevelSuffix(self, level):
        "Returns file name suffix accoding to backup level."

        levelSuffix = ""
        if options.incremental and level > 0:
            levelSuffix = "." + str(level)
        return levelSuffix



    def _getLevelFromFileName(self, fileName, suffix):
        """Extracts backup level number from file name.

        fileName should be in form '<self.archiveSpec.name>.[<level>]<suffix>'.
        """
        
        level = 0

        # partition file by self.archiveSpec.name first so we should get
        # level + suffix in last item or only suffix without leading '.'
        # in case that level was not present; that result partition again
        # by suffix
        partitioned = fileName.partition(
            self.archiveSpec.name + ".")[2].partition(suffix)

        # if partitioned[1] is not empty it means that suffix was present in
        # previous (second) partitioning so level is also present
        if partitioned[1]:
            level = int(partitioned[0])

        return level



    def _getMaxLevel(self):
        "Returns maximal backup level that can be created."

        return 0



    def _getErrorMessage(self, processReturnCode):
        """Returns error message based on processReturnCode.

        processReturnCode is exit code of the archiver process."""

        return ""



    def _getErrorCode(self, processReturnCode):
        """Returns ErrorCodes value based on processReturnCode.

        Base implementation directly maps processReturnCode to ErrorCodes
        values or to ErrorCodes.Error if higher."""

        if processReturnCode >= 0 and processReturnCode < ErrorCodes.count:
            errorCode = processReturnCode
        else:
            errorCode = ErrorCodes.Error

        return errorCode



    def __getSuffix(self):
        return self._SUFFIX

    suffix = property(__getSuffix, doc = "Gets archive suffix.")



    def __getLevel(self):
        "Returns level for new archive."

        level = self._getMaxLevel()

        if options.level != None:
            if options.level > level:
                raise ValueError, "Backup level value too high."
            level = options.level
        else:
            level = self._getMaxLevel()

        return level



class TarArchiver(Archiver):
    "Represents GNU tar archiver."

    ARCHIVER = "/bin/tar"
    _SUFFIX = ".tar"

    # suffix for snapshot files used in incremental backups
    SNAPSHOT_SUFFIX = ".snar"

    def execute(self):
        "See Archiver.execute."
        
        if options.incremental:

            if self.level > 0:

                # copy old snapshot file to new one
                shutil.copyfile(self.__getSnapshotFileName(self.level - 1),
                                self.__getSnapshotFileName(self.level))

            else:

                # we are going to create level 0 backup so remove snapshot
                # for level 0 because new one will be created
                if (os.path.exists(self.__getSnapshotFileName(0))):
                        os.remove(self.__getSnapshotFileName(0))

            # remove higher level snapshot files
            for snapshot in self.__getSnapshots():
                if self._getLevelFromFileName(snapshot, self.SNAPSHOT_SUFFIX) >\
                        self.level:
                    os.remove(os.path.join(createSnapshotsDir(), snapshot))

        super(TarArchiver, self).execute()



    def arguments(self):
        "See Archiver.arguments."

        # operation has to be first one
        archiverOptions = ["--create"]

        # add converted options passed to aa
        archiverOptions += self.__convertOptions()

        # add options required for this archiver
        archiverOptions += ["--file=" + self.archiveFileName, "--directory=" +
                            self.archiveSpec.path]

        # add converted include and exclude files
        archiverOptions += self.__convertArchiveSpec()

        return (archiverOptions, {})



    def _getMaxLevel(self):
        "See Archiver.getMaxLevel."

        level = 0

        # get list of snapshot files
        snapshotFiles = self.__getSnapshots()

        # if at least one snapshot file was found it means that level will
        # be > 1 (even if the 'level from file name extraction' would not be
        # successfull because level 0 file has no level suffix)
        if len(snapshotFiles) > 0:

            # find highest level
            for snapshotFile in snapshotFiles:
                newLevel = self._getLevelFromFileName(
                    snapshotFile,
                    self.SNAPSHOT_SUFFIX)
                if newLevel > level:
                    level = newLevel

            level += 1

        return level
        


    def __convertOptions(self):
        """Returns list of options for archiver.

        This basically converts options passed to aa to similar
        options for archiver program.
        """

        archiverOptions = []

        if options.verbose:
            archiverOptions.append("--verbose")

        if options.incremental:
            archiverOptions.append("--listed-incremental=" +
                                   self.__getSnapshotFileName(self.level))

        return archiverOptions



    def __convertArchiveSpec(self):
        """Returns list of arguments for archiver.

        Converts list of files and list of excluded files to the form suitable
        for archiver program.
        """

        archiverOptions = []

        for exclude in self.archiveSpec.excludesExpanded:
            archiverOptions.append("--exclude=" + exclude)

        archiverOptions += self.archiveSpec.includesExpanded

        return archiverOptions



    def __getSnapshotFileName(self, level):
        "Returns full path to shnapshot file for certain backup level."

        return os.path.join(createSnapshotsDir(), self.archiveSpec.name +
                            self._getLevelSuffix(level) +
                            self.SNAPSHOT_SUFFIX)



    def __getSnapshots(self):
        "Returns list of snapshot file names."

        snapshots = glob.glob(os.path.join(createSnapshotsDir(),
                                           self.archiveSpec.name + "*" +
                                           self.SNAPSHOT_SUFFIX))
        return [os.path.basename(snapshot) for snapshot in snapshots]



class TarCompressedArchiver(TarArchiver):
    "Represents GNU tar archiver with a compression."

    # suffix specific for compressed tar archive
    COMPRESS_SUFFIX = None

    # tar option for creating specific commpresed archive
    COMPRESS_OPTION = None

    # name of environment variable from which the compress program reads options
    ENV_NAME = None

    def __init__(self, archiveSpec):

        # raise exception because this class is meant as abstract
        if not self.COMPRESS_SUFFIX and not self.COMPRESS_OPTION:
            raise NotImplementedError

        super(TarCompressedArchiver, self).__init__(archiveSpec)



    def arguments(self):
        "See Archiver.arguments."

        # get arguments from base class
        baseOptions, baseEnvironment = \
            super(TarCompressedArchiver, self).arguments()

        # convert options passed to aa to arguments
        archiverOptions, environment = self.__convertOptions()

        # insert options required for this archiver
        archiverOptions.append(self.COMPRESS_OPTION)

        # assemble options so first item from baseOptions remains first
        # because it is the operation
        archiverOptions = baseOptions[:1] + archiverOptions + baseOptions[1:]

        # add environment from base class
        environment.update(baseEnvironment)

        return (archiverOptions, environment)



    def __getSuffix(self):
        return super(TarCompressedArchiver, self).suffix + self.COMPRESS_SUFFIX

    suffix = property(__getSuffix, doc = "See Archiver.suffix.")



    def __convertOptions(self):
        """Returns tuple of options and environment for archiver.

        This basically converts options passed to aa to similar
        options and environment for archiver program.
        """
        
        environment = {}

        if options.compresion_level != None and self.ENV_NAME:
            environment[self.ENV_NAME] = "-" + options.compresion_level

        return ([], environment)



class TargzArchiver(TarCompressedArchiver):
    "Represents GNU tar archiver with gzip compression."

    COMPRESS_SUFFIX = ".gz"
    COMPRESS_OPTION = "--gzip"
    ENV_NAME = "GZIP"



class TarbzArchiver(TarCompressedArchiver):
    "Represents GNU tar archiver with bzip2 compression."

    COMPRESS_SUFFIX = ".bz2"
    COMPRESS_OPTION = "--bzip2"
    ENV_NAME = "BZIP2"



class TarlzmaArchiver(TarCompressedArchiver):
    "Represents GNU tar archiver with lzma compression."

    COMPRESS_SUFFIX = ".lzma"
    COMPRESS_OPTION = "--lzma"



class ExecuteException(Exception):
    "Raised when some error ocurrs during archiver program execution."

    def __init__(self, message = "", errorCode = ErrorCodes.Undefined, *args):
        self.__message = message
        self.errorCode = errorCode
        super(ExecuteException, self).__init__(message, errorCode, *args)



    def __str__(self):
        return self.__message
